/* P-1 factoring code, ver. 0.1 (Feb 3 2000) by Alexander Kruppa, 
   alexander.kruppa@stud.tu-muenchen.de */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include "gmp.h"
#include "pminus1.h"

mpz_t N;		/* the number to factor */
mpz_t f;		/* the factor we found */
mpz_t extraexponent;	/* Extra number to exponentiate by */
char readline[256];
unsigned long bound1, bound2;
int 	argidx=1;
int	param=0;
int	mersenne=0;

void printUsage() {
  printf("Usage: gpm1 [-v] [-h] [-m] <number> <limit> <extraexponent>\n");
  printf(" -v  be verbose (print primes as they are processed)\n");
  printf(" -h  print this help\n");
  printf(" -m  exponentiate by <number>-1 (for Mersennes etc)\n");
  printf("     If <number> is - , then the number will be read from stdin\n");
}


int main(int argc, char *argv[]){
  mpz_init(N);
  mpz_init(f);
  mpz_init(extraexponent);
  mpz_set_ui(extraexponent,1);

  if (argc < 3) {
    printUsage();
    return 1;
  }

  param += PM1DOOUTPUT;

nextparam:
  if (strncmp(argv[argidx],"-v",2) == 0) {
    argidx++;
    param += PM1PROGRESS;
    goto nextparam;
  }
  
  if (strncmp(argv[argidx],"-m",2) == 0) {
    mersenne = 1;
    argidx++;
    goto nextparam;
  }
  
  if (strncmp(argv[argidx],"-h",2) == 0) {
    printUsage();
    return 0;
  }

  if (strncmp(argv[argidx],"-",1) == 0) {    
    mpz_inp_str(N, stdin, 10);
  } else {
    mpz_set_str(N,argv[argidx],10);
  }
  argidx++;

  if (mpz_cmp_ui(N,0) == 0) {
    printf("You must give an integer number to factor!\n");
    printUsage();
    return 1;
  }

  if (argv[argidx] == 0) {
    printf("You must specify a bound value!\n");
    printUsage();
    return 1;
  }
  bound1 = atol(argv[argidx++]);
  bound2 = bound1 * 40;

  if (argc > argidx ) {
    mpz_set_str(extraexponent, argv[argidx++], 10);
    if (mpz_cmp_ui(extraexponent,0) == 0) {
      printf("The extra exponent must be an integer number!\n");
      printUsage();
      return 1;
    }
  }

  if (mersenne) {
    mpz_sub_ui(N,N,1);				/* !! What if N=1 ? */
    mpz_mul(extraexponent, extraexponent, N);	/* For Mersenne numbers etc */
    mpz_add_ui(N,N,1);
  }

  if ( pminus1(f, bound1, bound2, N, extraexponent, param) ) {
/*    mpz_divexact(N,N,f);
    printf("Cofactor: ");
    mpz_out_str(stdout,10,N);
    printf("\n");
*/  }

  return 0;
}
