/* P-1 factoring code, ver. 0.3 (Jan 23 2000) by Alexander Kruppa, 
   alexander.kruppa@stud.tu-muenchen.de 
   2000 Feb. 7:	Fixed bug in stage 2, x was not initialised correctly which turned stage 2 
   		mostly useless
   2000 Jun. 2:	Sieve only stores odd numbers; number of even-powers-of-stage-1-residue to 
   		use is now taken from list, not computed anymore - should save a little memory
   2000 Aug. 3: E in stage 1 computed explicitly for low bounds, improved stage 2 (twice as 
   		fast, but uses more memory)
   2001 Jan.23: mpz_ui_powm() to exponentiate a unsigned long by a huge number.
   		Uses only mpz_mul_ui() and squarings. Speed is almost the same :(
*/

#include "pminus1.h"
#include <stdlib.h>
#include <primegen.h>

/* For prime sieve */
#define SIEVE_NUMBER(i) (2*i+1)
#define SIEVE_INDEX(n) ((n-1)/2)

void mpz_ui_powm(mpz_t r, unsigned long base, mpz_t exp, mpz_t m) {

  unsigned int expidx;
  mp_limb_t bitmask, expbits;

  expidx = mpz_size (exp) -1;		/* point at most significant limb */
  expbits = mpz_getlimbn (exp, expidx);	/* get most significant limb */
  bitmask = 1<<(sizeof(mp_limb_t)*8-1);

  while ((bitmask & expbits) == 0) {
    bitmask >>= 1;
    if (bitmask == 0) {			/* no set bits in this limb */
      if (expidx == 0) {		/* no more limbs -> exp was 0 */
        mpz_set_ui(r, 1);		/* set result to 1 and return */
        return;
      }
      expidx--;
      expbits = mpz_getlimbn (exp, expidx);
      bitmask = 1<<(sizeof(mp_limb_t)*8-1);
    }
  }

/* here the most significant limb with any set bits is in expbits, */
/* bitmask is set at the msb of expbits */

  mpz_set_ui (r, base);
  bitmask >>= 1;

  while (1) {
    while (bitmask != 0) {
      mpz_mul(r, r, r);			/* r = r^2 */
      mpz_mod(r, r, m);
      if (expbits & bitmask) mpz_mul_ui (r, r, base);
      bitmask >>= 1;
    }
    if (expidx == 0)			/* if we just processed the least */
      break;				/* significant limb, we are done */
    expidx--;
    expbits = mpz_getlimbn (exp, expidx);
    bitmask = 1<<(sizeof(mp_limb_t)*8-1);
  }
  mpz_mod(r, r, m);
  return;
}
  
int pminus1(mpz_t f,unsigned long bound1,unsigned long bound2,mpz_t N, mpz_t extraexponent, int param) 
{
#define ROOT 3
#define PRIMEGAPC 34
  unsigned long primegapn[PRIMEGAPC]=	{2,3,5,11,29,97,127,541,907,1151,1361,9587,15727,19661,31469,156007,360749,370373,492227,1349651,1357333,2010881,4652507,17051887,20831533,47326913,122164969,189695893,191913031,387096383,436273291,1294268779,1453168433,2300942869LU};
  int primegapi[PRIMEGAPC]=		{0,1,2, 4, 6, 8, 14, 18, 20,  22,  34,  36,   44,   52,   72,    86,    96,   112,   114,    118,    132,    148,    154,     180,     210,     220,      222,      234,      248,      250,      282,       288,       292,       320};

  mpz_t x;		/* 3^E mod N where E is product of all primes<bound */
  mpz_t E;
  mpz_t m,d;
  mpz_t xpowD;		/* x^D */
  mpz_t *s2res;		/* Stored residues for stage 2 */
  unsigned long i,j,	/* loop counters */
	prime,		/* the current prime as a longint */
	lastprime,	/* for stage 2: the last prime processed */
	primediff,	/* difference: lastprime-prime */
	s2size,		/* How many residues we store for stage 2 */
	lpow,		/* the power we put a prime into as a longint */
        sqrtbound,	/* the square root of the bound2 */
        D,		/* sqrt(B2-B1) */
        u;
  primegen pg[1];

#ifdef DEBUG
  unsigned long nrprimes=0;	/* The number of primes we processed */
#endif
  double logbound;	/* the logarithm of the bound2 */
  char *sieve;
  int 	doOutput=0,
  	printProgress=0,
  	explicitE=0;

#ifdef DEBUG
  printf("Called with B1=%lu, B2= %lu, param=%lu\n");
#endif

#ifdef DEBUG
  /* Test the list of prime gaps */
  {
    mpz_t p;
    mpz_init(p);
    for (i=0; i<PRIMEGAPC; i++) {
      mpz_set_ui(p,primegapn[i]);
      if (!mpz_probab_prime_p(p,20)) {
        fprintf(stderr,"Internal error, %lu is not prime!\n",primegapn[i]); 
        return 0;
      }
      mpz_set_ui(p,primegapn[i]-primegapi[i]);
      if (!mpz_probab_prime_p(p,20)) {
        fprintf(stderr,"Internal error, %lu is not prime!\n",primegapn[i]-primegapi[i]);
        return 0;
      }
      for (j=primegapn[i]-primegapi[i]+1; j<primegapn[i]; j++) {
        mpz_set_ui(p,j);
        if (mpz_probab_prime_p(p,20)) {
          fprintf(stderr,"Internal error, %lu, between %lu and %lu, is a prime!\n",j,primegapn[i]-primegapi[i],primegapn[i]);
          return 0;
        }
      }
    }
  }
#endif

  if (bound2 < bound1) bound2 = bound1;
  sqrtbound = (unsigned long) sqrt(bound2);
  logbound = log(bound2);

  if (param & PM1DOOUTPUT) doOutput=1;
  if (param & PM1PROGRESS) printProgress=1;
  if ((bound1 <= 100000) && ((param & PM1NOEXPLICITE) == 0)) {
    explicitE = 1;
#ifdef DEBUG
    printf("Computing E explicitly\n");
#endif
  }    


  mpz_init_set_ui(x, ROOT);	/* x = 3 */
  if (explicitE) mpz_init_set_ui(E, 1);

  primegen_init(pg);

  if (doOutput) printf("Stage 1 ");
  fflush(stdout);

  /* Exponentiate by extra exponent */
  if (!explicitE) mpz_powm(x,x,extraexponent,N);
             else mpz_mul(E,E,extraexponent);

  prime=2;
  i=0;		/* Points at 1 now */

  /* Process all primes which we put into a power > 1 */
  while(prime<=sqrtbound) {
    lpow=(int)(logbound/log(prime));	/* Compute power to put prime into */
    if (printProgress) {
      printf("%lu^%lu*",prime,lpow); 
      fflush(stdout); 
    }
    if (!explicitE) for (j=0; j<lpow; j++) mpz_powm_ui(x,x,prime,N);
    else {
      unsigned long t = 1;
      for (j=0; j<lpow; j++) t *= prime;
      mpz_mul_ui(E, E, t);
    }
      
#ifdef DEBUG
    nrprimes++;
#endif
    prime = primegen_next(pg);
  }

  while(prime<=bound1) {
    if (printProgress) {
      printf("%lu*",prime);
      fflush(stdout);
    }
    if (!explicitE) mpz_powm_ui(x,x,prime,N);
               else mpz_mul_ui(E, E, prime);
#ifdef DEBUG
    nrprimes++;
#endif    
    prime = primegen_next(pg);
  }

  /* prime > bound1 here */

  if (explicitE) {
    mpz_ui_powm(x,ROOT,E,N);
    mpz_clear(E);
  }

  /* If we found a factor, print it (if desired) and return */
  mpz_sub_ui(f,x,1);			/* f = x-1 */
  mpz_gcd(f,f,N);			/* f = gcd(f, N) */
  if (mpz_cmp_ui(f,1) > 0) {		/* if f>1 */
    if (doOutput) { 
      printf(" factor found: ");
      mpz_out_str(stdout,10,f);
      printf("\n");
    }
    mpz_clear(x);
    free(sieve);
    return 1;				/* Return, signal factor in stage 1 */
  }

  if (doOutput) printf("\nStage 2 ");
  fflush(stdout);

  /* Stage 2 */

  D = (unsigned long) sqrt(bound2-bound1) & ~1UL;
  s2size = D/2;
  s2res = malloc(s2size * sizeof(mpz_t));
  mpz_init(xpowD);
  mpz_powm_ui(xpowD, x, 2, N);	/* xpowD = x^2 for now */
  mpz_init_set_ui(s2res[0],1);
  for (j=1; j<s2size; j++) {
    mpz_init(s2res[j]);
    mpz_mul(s2res[j], s2res[j-1], xpowD);
    mpz_mod(s2res[j], s2res[j], N);
  }
  /* Now s2res[i] = x^(2*i) */
  mpz_powm_ui(xpowD, x, D, N);	/* xpowD = x^D */
  u = prime;
  mpz_init(m);
  mpz_powm_ui(m, x, prime, N);
  mpz_set_ui(f, 1);
  mpz_init(d);
  
  while (prime <= bound2) {

    if (printProgress) {
      printf("%lu ",prime);
      fflush(stdout);
    }
    if (prime > u) {
      u = u + D;
      mpz_mul(m, m, xpowD);
      mpz_mod(m, m, N);
    }

#ifdef DEBUG
    nrprimes++;
    if ( (u-prime) >= 2*s2size) {
      printf("(u-prime) >= 2*s2size !");
      exit(EXIT_FAILURE);
    }
#endif
      
    mpz_sub(d, m, s2res[(u - prime)/2]);
    if (mpz_sgn(d) == -1) mpz_add(d,d,N);
    mpz_mul(f,f,d);
    mpz_mod(f,f,N);

    prime = primegen_next(pg);
  }

  /* Free up memory */
  mpz_clear(m);
  mpz_clear(d);
  mpz_clear(xpowD);
  for (j=0; j<s2size; j++) mpz_clear(s2res[j]);
  free(s2res);
  
  mpz_gcd(f,f,N);
  if (mpz_cmp_ui(f,1) > 0) {		/* if f>1 */
    if (doOutput) {
      printf(" factor found: ");
      mpz_out_str(stdout,10,f);
      printf("\n");
    }
    return 2;				/* Signal factor in stage 2 */
  }
  if (doOutput) printf("\n");
#ifdef DEBUG
  printf("Number of primes processed: %lu\n",nrprimes);
#endif
  return 0;				/* No factor found */
}
